open Client
module Log = Test.Log

(* ∀a b. ({} -> (a * (a -> b))) -> b
   Λa b. fun (p : {} -> (a * (a -> b))) ->
     let (x, f) = p () in f x
*)
let let_prod =
  let open F in
  let alpha, beta = 0, 1 in
  let p, f, x = "p", "f", "x" in
  TyAbs (alpha, TyAbs (beta,
    Abs (p, TyArrow (TyProduct [],
                     TyProduct [TyVar alpha; TyArrow (TyVar alpha, TyVar beta)]),
      LetProd ([x; f], App (Var p, Tuple []),
        App (Var f, Var x)))))

(* Inj<∀α. α → α + ∀βα. α → β> 0 Λα.λx:α.x *)
let variant =
  let alpha, beta = 0, 1 in
  F.(
    Inj
      (
        [
          TyForall (alpha, TyArrow (TyVar alpha, TyVar alpha)) ;
          TyForall (beta, TyForall (alpha, TyArrow (TyVar alpha, TyVar beta)))
        ],
        0,
        TyAbs (alpha, Abs("x", TyVar alpha, Var "x"))
      )
  )

(* Λαβ. λx. match x with (_, y) -> y end *)
let match_with =
  let alpha, beta = 1, 0 in
  F.(
    TyAbs(alpha,
      TyAbs(beta,
        Abs("x",
          TyProduct [ TyVar alpha ; TyVar beta ],
          Match(TyVar beta,
                Var "x",
                [
                  (PTuple [ PWildcard ; PVar "y" ] ,
                   Var "y")
                ]
            )
          )
        )
      )
  )

let print_term t =
  PPrint.(ToChannel.pretty 0.9 80 stdout
            (FPrinter.print_term t ^^ hardline))

let test_f (t : F.nominal_term) =
  let log = Log.create_log () in
  print_term t;
  let t : F.debruijn_term =
    Log.attempt log
      "Converting the System F term to de Bruijn style...\n"
      F.translate t
  in
  let _ty : F.debruijn_type =
    Log.attempt log
      "Type-checking the System F term...\n"
      FTypeChecker.typeof t
  in
  ()

let () =
  test_f let_prod;
  test_f variant;
  test_f match_with;
  ()
