open Client

(* A few manually constructed terms. *)

let x =
  ML.Var "x"

let y =
  ML.Var "y"

let id =
  ML.Abs ("x", x)

let delta =
  ML.Abs ("x", ML.App (x, x))

let deltadelta =
  ML.App (delta, delta)

let idid =
  ML.App (id, id)

let k =
  ML.Abs ("x", ML.Abs ("y", x))

let genid =
  ML.Let ("x", id, x)

let genidid =
  ML.Let ("x", id, ML.App (x, x))

let genkidid =
  ML.Let ("x", ML.App (k, id), ML.App (x, id))

let genkidid2 =
  ML.Let ("x", ML.App (ML.App (k, id), id), x)

let app_pair = (* ill-typed *)
  ML.App (ML.Tuple [id; id], id)

(* option *)
let option_env =
  
  (* type 'a option = None | Some of 'a *)
  let option_typedecl = ML.{
    name = Type "option";
    type_params = [ "a" ];
    data_kind = true;
    labels_descr = [
      {
        label_name = Label"None";
        type_name = Type "option";
        arg_type = TyProduct [];          
      } ; {
        label_name = Label "Some";
        type_name = Type "option";
        arg_type = TyVar "a";
      }
    ]
  } in

  Infer.Env.add_typedecl Infer.Env.empty option_typedecl
  
let none = ML.Variant ( ML.Label "None" , ML.Tuple [] )

let some =
  ML.Variant (
    ML.Label "Some",
    ML.Abs ("x", ML.Var "x")
  )

(* list *)
let list_env =

  (* type 'a list = Nil | Cons of 'a * 'a list *)
  let list_typedecl = ML.{
    name = Type "list";
    type_params = [ "a" ];
    data_kind = true;
    labels_descr = [
      {
        label_name = Label "Nil";
        type_name = Type "list";
        arg_type = TyProduct [];          
      } ; {
        label_name = Label "Cons";
        type_name = Type "list";
        arg_type = TyProduct [ TyVar "a" ; TyConstr (Type "list", [ TyVar "a" ]) ];
      }
    ]
  } in

  Infer.Env.add_typedecl option_env list_typedecl

let nil = ML.Variant ( ML.Label "Nil" , ML.Tuple [] )

let list =
  ML.Variant (
    ML.Label "Cons",
    ML.Tuple [ ML.Abs ("x", ML.Var "x") ; nil ]
  )

(* tree *)
let tree_env =

  (* type 'a tree = Leaf | Node of 'a tree * 'a * 'a tree *)
  let tree_typedecl = ML.{
    name = Type "tree";
    type_params = [ "a" ];
    data_kind = true;
    labels_descr = [
      {
        label_name = Label "Leaf";
        type_name = Type "tree";
        arg_type = TyProduct []
      } ; {
        label_name = Label "Node";
        type_name = Type "tree";
        arg_type = TyProduct [
          TyConstr (Type "tree", [ TyVar "a" ]);
          TyVar "a";
          TyConstr (Type "tree", [ TyVar "a" ]);
        ]
      }
    ]
  } in

  Infer.Env.add_typedecl list_env tree_typedecl

let leaf = ML.Variant ( ML.Label "Leaf" , ML.Tuple [] )

let node =
  ML.Variant (
    ML.Label "Node",
    ML.Tuple [
      leaf ;
      ML.Abs ("x", ML.Var "x") ;
      leaf ;
    ]
  )

let () =
  assert (Test.CheckML.test Infer.Env.empty idid);
  assert (Test.CheckML.test Infer.Env.empty genid);
  assert (Test.CheckML.test Infer.Env.empty genidid);
  assert (Test.CheckML.test Infer.Env.empty genkidid);
  assert (Test.CheckML.test Infer.Env.empty genkidid2);
  assert (Test.CheckML.test option_env none);
  assert (Test.CheckML.test option_env some);
  assert (Test.CheckML.test list_env nil);
  assert (Test.CheckML.test list_env list);
  assert (Test.CheckML.test tree_env leaf);
  assert (Test.CheckML.test tree_env node);
  (* we include some printing below because Dune
     appears to only show tests that show some output. *)
  print_endline "TestML: all tests passed."
